/*
 * XLife Copyright 2013 Vladimir Lidovski vol.litwr@gmail.com
 * $Id$
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "defs.h"
#include "framebuffer.h"

struct FB *first, *pfb;

void fb_ins(coord_t xa, coord_t ya, cell_t c) {
   register struct FB *p;
   if (xa < xpos || ya < ypos)
      return;
   p = pfb + maxwidth*shl(ya - ypos, scale) + shl(xa - xpos, scale);
   if (c && !p->is) {
      p->old = 0;
      p->is = 1;
      if (first)
         p->next = first;
      else
         p->next = 0;
      first = p;
   }
   if (p->is) {
      p->cur = c;
      p->checked = 1;
   }
}

void fb_ins_old(coord_t xa, coord_t ya, cell_t c) {
   register struct FB *p;
   if (xa < xpos || ya < ypos)
      return;
   p = pfb + maxwidth*shl(ya - ypos, scale) + shl(xa - xpos, scale);
   if (!p->is && c) {
      p->cur = 0;
      p->is = 1;
      if (first)
         p->next = first;
      else
         p->next = 0;
      first = p;
   }
   if (p->is) {
      p->old = c;
      p->checked = 1;
   }
}

static inline void putcell(struct FB *p, cell_t c) {
   unsigned xy = p - pfb;
   unsigned x = xy%maxwidth, y = xy/maxwidth;
   if (scale > 0) {
      rects[c][chgrects[c]].x = x;
      rects[c][chgrects[c]++].y = y;
      if (chgrects[c] == MAXCHANGE) {
         XFillRectangles(disp, lifew, cellgc[c], rects[c], MAXCHANGE);
         chgrects[c] = 0;
      }
   }
   else {
      points[c][chgpoints[c]].y = y;
      points[c][chgpoints[c]++].x = x;
      if (chgpoints[c] == MAXCHANGE) {
         XDrawPoints(disp, lifew, cellgc[c], points[c], MAXCHANGE, CoordModeOrigin);
         chgpoints[c] = 0;
      }
   }
}

void fb_flush() {
   register struct FB *p, *po;
   for (p = first; p; p = p->next)
      if (p->cur == 0 && p->old == 0) {
         if (p == first)
           first = p->next;
         else
           po->next = p->next;
         p->is = 0;
      }
      else if (p->checked == 0) {
         if (p->cur) putcell(p, 0);
         if (p == first)
           first = p->next;
         else
           po->next = p->next;
         p->is = 0;
      }
      else {
         p->checked = 0;
         po = p;
         if (p->old != p->cur || force_fb) {
            putcell(p, p->cur);
            p->old = p->cur;
         }
      }
}

void fb_clear() {
   register struct FB *p;
   for (p = first; p; p = p->next)
      p->is = 0;
   first = 0;
}
