/*
 * XLife Copyright 2011, 2012 Vladimir Lidovski vol.litwr@gmail.com
 * $Id: map.cpp 78 2012-05-16 13:10:01Z micro $
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <map>
#include "defs.h"
#include "tile.h"
#define MAXOSCHISTORY 200000
#define MAX_OSC_RESULTS 2000
using namespace std;

struct OSCID {
   cellcount_t         tilecount;
   cellcount_t         chgcount;
   cellcount_t         cellcount[MAXSTATES];
   unsigned long       generations;
};

unsigned long map_results[MAX_OSC_RESULTS];
static multimap<cellcount_t, OSCID> testpats;

static cellcount_t acellcnt() {
   cellcount_t sum = 0;
   for (unsigned i = 1; i < maxstates; i++)
      sum += active.cellcount[i];
   return sum;
}

static inline cellcount_t calckey(void) {
   return (active.tilecount << 24) ^ (active.chgcount << 12) ^ acellcnt();
}

extern "C" int map_info() {
   return testpats.size();
}

extern "C" void map_init() {
   testpats.clear();
}

extern "C" int map_add(void) {
   if (active.cellcount > 0)
      testpats.insert(make_pair(calckey(), *(OSCID*)&active.tilecount));
   if (testpats.size() >  MAXOSCHISTORY) {
      state = STOP;
      testpats.clear();
      return 1;
   }
   return 0;
}

extern "C" int map_find(void) {
   typedef multimap<cellcount_t, OSCID>::const_iterator I;
   pair<I, I> g = testpats.equal_range(calckey());
   int q = 0;
   for (I p = g.first; p != g.second; ++p)
      if (active.tilecount == p->second.tilecount && active.chgcount == p->second.chgcount)
         if (!memcmp(p->second.cellcount + 1, active.cellcount + 1, maxstates - 1))
            if (q < MAX_OSC_RESULTS)
               map_results[q++] = p->second.generations;
   return q;
}
